/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *
 **/

#define _GNU_SOURCE

#define DEBUG_PK 0

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "socket_srv.h"
#include "iar_bus.h"
#include "my_stdio.h"


static struct ib_dev *init_ib_dev()
{
	struct ib_dev *dev;

	if ((dev = (struct ib_dev *) malloc(sizeof(struct ib_dev))) == NULL) {
		perror("init_ib_dev");
		exit(1);
	}
       	dev->fd = -1;
	dev->addr = IB_NULL_ADDR;
	dev->status = IB_DEV_RW;
	dev->status &= ~ IB_DEV_ALIVE;
	dev->next = dev->prev = NULL;
	
	return dev;
}


static int set_ib_dev_addr(struct ib_dev *dev, int addr)
{
	if (addr == 0)	
		dev->addr = IB_BASE_ADDR | dev->fd ;
	else 
		dev->addr = addr;

	return 0;
}


static int free_ib_dev(struct ib_dev *dev)
{
	if( dev->fd >= 0 )
		close(dev->fd);
	free(dev);

	return 0;
}


static struct ib_dev *find_ib_dev (struct ib_dev *dev, int addr)
{
	struct ib_dev *devp = dev;
	
	while (devp != NULL ){
		if (devp->addr == addr) {
			return devp;
		}
		devp = devp->next;
	}

	return NULL;
}


static int die_ib_dev(struct ib_dev *devp)
{
	PVERB(PWARN, "Core bus: read/write device fail\n");
	close(devp->fd);
	devp->fd = -1;
	devp->status &= ~ IB_DEV_ALIVE;

	return 0;
}


static int send_data_to_dev(struct ib_packet *d, int to_fd)
{
	if( write(to_fd, d, sizeof(struct ib_packet_head)) <= 0 ){
		return -1;
	}
	if( d->h.size > 0 ) {
		if( write(to_fd, d->data, d->h.size) < 0 ){
			return -1;
		}
	}

	return 0;
}


static int send_to_dev(struct ib_dev *from, struct ib_dev *to)
{
	if( to->status & (IB_DEV_ON | IB_DEV_ALIVE | IB_DEV_READ) ) {
		from->pk.h.addr = from->addr;
#if DEBUG_PK
		PVERB(PINFO, "Core bus: send from 0x%x to 0x%x, size %d\n",
		      from->pk.h.addr, to->pk.h.addr, from->pk.h.size);
#endif
		if( send_data_to_dev( &from->pk, to->fd) < 0 ){
			die_ib_dev( to );
			return -1;
		}
	}

	return 0;
}


static struct ib_command *find_ib_command(struct ib_command *cm, int cm_id)
{
	int i;

	for (i = 0; cm[i].func; i++) {
		if (cm[i].cm_id == cm_id ) {
			return (&cm[i]);
		}
	}
	
	return ((struct ib_command *)NULL);
}


/* Execute a command. */
static int execute_command(struct ib_command *cm, struct ib_dev *dev,
			   struct ib_dev *devp)
{
	struct ib_command *command;
	
	command = find_ib_command (cm, devp->pk.h.cmd);

	if (!command) {
		return -1;
	}

	/* Call the function. */
	return ((*(command->func)) (dev, devp));
}


static int read_cm_from_dev(struct ib_dev *devp)
{
	memset(&devp->pk, 0, sizeof(struct ib_packet_head));
	if ( read(devp->fd, &devp->pk.h, sizeof(struct ib_packet_head)) <= 0) {
		/* Ops... die */
		die_ib_dev( devp );
		return -1;
	}
	if (  devp->pk.h.size > 0 && devp->pk.h.size < IB_PACKET_DATA_LEN ){
		memset(devp->pk.data, 0, devp->pk.h.size);
		if ( read(devp->fd, devp->pk.data, devp->pk.h.size) <= 0 ) {
			die_ib_dev( devp );
			return -1;
		}
	}
	devp->status |= IB_DEV_NEWPK;
#if DEBUG_PK 
	  PVERB(PINFO, "Core bus: rcv from 0x%x to 0x%x, size %d\n",
	  devp->addr, devp->pk.h.addr, devp->pk.h.size);
#endif
	return 0;
}


static int process_cm_queue(struct ib_command *cm, struct ib_dev *dev)
{
	struct ib_dev *devp = dev;

	while(devp != NULL) {
		if ( devp->status & IB_DEV_NEWPK ){
			execute_command(cm, dev, devp);
			devp->status &= ~IB_DEV_NEWPK;
		}
		devp = devp->next;
	}

	return 0;
}


void intHandler (int signo) 
{
	PVERB(PERR, "Core bus: write error, SIGPIPE trapped\n");
	return;
}


static int iar_bus_core(struct ib_command *cm, int fd_socket )
{
	struct ib_dev *dev, *devp, *devd;
	fd_set rfds;
	struct timeval tv;
	int retval;
	struct sockaddr isa;
	int addrlen;
	int queue;
	int fd_max;

	addrlen = sizeof(isa); /* find socket's address for accept() */
	getsockname(fd_socket, (struct sockaddr *)&isa, &addrlen);
	if (signal(SIGPIPE, intHandler) == SIG_ERR) {
		perror("Signal Failed (SIGPIPE).");
		exit(1);
	}
	dev = init_ib_dev();
	/* bus dev */
	dev->fd = fd_socket;
	dev->addr = IB_BUS_ADDR;
	dev->status = IB_DEV_ON | IB_DEV_ALIVE | IB_DEV_READ;

	for(;1;){
		devp = dev;
		FD_ZERO(&rfds);
		fd_max = fd_socket + 1;
		while(devp != NULL) {
			if (devp->fd && (devp->status & IB_DEV_ALIVE) ){ 
				FD_SET(devp->fd, &rfds);
				if( fd_max < devp->fd + 1)
					fd_max = devp->fd + 1;
			}
			else if ( !(devp->status & IB_DEV_ALIVE) && 
				  (devp != dev) ) { 
                                /* OK: device dead, cleaning */ 
				devd = devp;
				devp = devd->prev;
				devp->next = devd->next;
				if( devd->next )
					devd->next->prev = devp;
				free_ib_dev(devd);	
			}
			devp = devp->next;
		}
		tv.tv_sec = 5;
		tv.tv_usec = 0;
		retval = select(fd_max, &rfds, NULL, NULL, &tv);
		if (retval) {
			devp = dev->next ; /* skip device bus */
			queue = 0;
			while(devp != NULL) {
				if(devp->status & IB_DEV_ALIVE)
					if( FD_ISSET(devp->fd, &rfds) ) {
						read_cm_from_dev(devp);
						queue = 1;
					}
				devp = devp->next;
			}
			if( queue )
				process_cm_queue(cm, dev->next);
			if (FD_ISSET(dev->fd, &rfds)) {
				/* New device is connected */
				PVERB(PINFO, "Core bus: new call\n");
				devp = dev ;
				while(devp->next != NULL) devp = devp->next;
				devp->next = init_ib_dev();
				devp->next->prev = devp;
				devp = devp->next;
				devp->fd = accept(dev->fd, 
						  (struct sockaddr *)&isa, 
						  &addrlen); 
				if( devp->fd ){
					devp->status |= IB_DEV_ALIVE;
					PVERB(PINFO, "Core bus: connected\n");
					set_ib_dev_addr(devp, 0);
				}
			}
		}
	}

	return 0;
}


/****************************************************************************
 * Start functions:
 **/

int init_iar_bus_local(struct ib_command *cm, char *sock_name )
{
	int fd_socket;

	if ( (fd_socket = socket_init_local( sock_name )) < 0) {
		return -1;
	}

	return iar_bus_core(cm, fd_socket );
}


int init_iar_bus_inet(struct ib_command *cm, int port )
{
	int fd_socket;

	if ( (fd_socket = socket_init( port )) < 0) {
		return -1;
	}

	return iar_bus_core(cm, fd_socket);
}


int iar_bus_start( char *sock_name )
{
	struct ib_command cm[] = {
		{IB_CM_MESSAGE, ib_cm_message},       /* tested */
		{IB_CM_MESSAGE_TO, ib_cm_message_to}, /* tested */
		{IB_CM_REMOVE,  ib_cm_remove},        /* untested */
		{IB_CM_ON,      ib_cm_on},            /* tested */
		{IB_CM_OFF,     ib_cm_off},           /* untested */
		{IB_CM_RW,      ib_cm_rw},            /* tested */
		//{IB_CM_SCAN, ib_cm_scan},           /* untested */
		//{IB_CM_WHOAMI, ib_cm_whoami},       /* untested */
		{IB_CM_SET_ADDR, ib_cm_set_addr},     /* tested */
		{-1, NULL}
	};

	return init_iar_bus_local( cm, sock_name );
}


/**
 * End Start functions.
 ****************************************************************************/


/****************************************************************************
 * Command functions:
 **/


int ib_cm_remove(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_packet d;

	d.h.addr = IB_BUS_ADDR;
	d.h.size = 0;
	d.h.cmd = IB_CM_REMOVE;

	send_data_to_dev(&d, current->fd);
	current->status &= ~ (IB_DEV_ON |IB_DEV_ALIVE) ;

	return close( current->fd );
}


int ib_cm_on(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_packet d;

	d.h.addr = IB_BUS_ADDR;
	d.h.size = 0;
	d.h.cmd = IB_CM_ON;

	current->status |= IB_DEV_ON;
	send_data_to_dev(&d, current->fd);

	return 0;
}


int ib_cm_off(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_packet d;

	d.h.addr = IB_BUS_ADDR;
	d.h.size = 0;
	d.h.cmd = IB_CM_OFF;

	send_data_to_dev(&d, current->fd);
	current->status &= ~ IB_DEV_ON;

	return 0;
}


int ib_cm_rw(struct ib_dev *dev, struct ib_dev *current)
{
	int new_state = ((int*) current->pk.data)[0];
	struct ib_packet d;

	new_state &= IB_DEV_RW;
	d.h.addr = IB_BUS_ADDR;
	d.h.size = sizeof(int);
	d.h.cmd = IB_CM_RW;
	memcpy(&d.data, &new_state, sizeof(int));
	current->status &= ~ IB_DEV_RW;
	current->status |= new_state;
	send_data_to_dev(&d, current->fd);

	return 0;
}


int ib_cm_message(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_dev *devp = dev;

	while (devp != NULL){
		if( (devp->status & IB_DEV_ALIVE) && (devp != current) ){
			send_to_dev(current, devp);
		}
		devp = devp->next;
	}
	
	return 0;
}


int ib_cm_message_to(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_dev *to = find_ib_dev(dev, current->pk.h.addr);

	if( to ){
		return send_to_dev(current, to );
	}
	
	return -1;
}


int ib_cm_set_addr(struct ib_dev *dev, struct ib_dev *current)
{
	int new_addr = ((int*) current->pk.data)[0];
	int ret;
	struct ib_packet d;

	d.h.addr = IB_BUS_ADDR;
	d.h.size = sizeof(int);
	d.h.cmd = IB_CM_SET_ADDR;

	memcpy(&d.data, &new_addr, sizeof(int));

	if( find_ib_dev(dev, new_addr) != NULL ||
		(new_addr & IB_BASE_ADDR)      ||
		(new_addr == IB_BUS_ADDR) ) {
		ret = -1;
		new_addr = current->addr;
	}else{
		ret = set_ib_dev_addr(current, new_addr);
	}
	send_data_to_dev(&d, current->fd);

	return ret;
}

/**
 * End command functions.
 ****************************************************************************/

/* End */
